'''
    Project: Gui Gin Rummy
    File name: player_type.py
    Author: William Hale
    Date created: 3/14/2020
'''

import enum


class PlayerType(enum.Enum):

    computer_player = 1
    human_player = 2
    demo_player = 3
